import argparse
import logging

# from pathlib import Path # TODOLATER:

import uvicorn

# Local imports
from .app import asgi_app, app  # noqa: F401

# Run Config
NUM_WORKERS = 1


if __name__ == "__main__":
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--port", type=int, default=80)
    parser.add_argument("-d", "--debug", action="store_true")  # False unless specified
    parser.add_argument("-l", "--debug_level", type=int, default=logging.DEBUG)
    parser.add_argument("-w", "--num_workers", type=int, default=NUM_WORKERS)
    parser.add_argument(
        "-s", "--use_dash_server", action="store_true"
    )  # False unless specified
    # parser.add_argument("-r", "--reset", type=bool, default=False)
    args = parser.parse_args()

    if args.debug:
        # Set env variable for Qimchi's debug level
        import os

        os.environ["QIMCHI_DEBUG_LEVEL"] = str(args.debug_level)

    # TODOLATER:
    # if args.reset:
    #     # Reset the state
    #     print("Resetting the state...")
    #     Path("state.json").unlink(missing_ok=True)

    if args.use_dash_server:
        logging.warning(
            "Qimchi is using Dash server as `-s` is set in the command. This is not recommended for production use."
            " Use Uvicorn instead."
        )
        app.run(
            host="0.0.0.0",
            port=args.port,
            debug=args.debug,  # Reload the server on changes, only in debug mode
            use_reloader=args.debug,  # Reload the server on changes, only in debug mode
        )
    else:
        uvicorn.run(
            "qimchi.__main__:asgi_app",
            host="0.0.0.0",
            port=args.port,
            log_level="error",  # NOTE: Separate from Qimchi's debug level
            reload=args.debug,  # Reload the server on changes, only in debug mode
            workers=args.num_workers,
        )



from pathlib import Path
from datetime import datetime
from string import Template

from dash import Input, Output, State, callback, html, dcc
from dash.exceptions import PreventUpdate

# Local imports
from qimchi.state import load_state_from_disk
from qimchi.logger import logger


# Default Frontmatter - Obsidian style
FRONTMATTER = Template(
    """---
Last Saved: ${timestamp}
Filename: ${filename}

---"""
)
DEFAULT_NOTES = "Type your notes here..."


def _get_notes(sess_id: str) -> str:
    """
    Utility function to load or create notes

    Args:
        sess_id (str): Session ID to load the state for

    Returns:
        str: Notes to be displayed in the notes viewer

    """
    _state = load_state_from_disk(sess_id)

    base = Path(_state.measurement_path)
    if not _state.measurement_last_fmt:
        logger.debug("Notes | No measurement found")
        return DEFAULT_NOTES

    path = base.parent / base.stem / (base.stem + ".md")

    if path.exists():
        logger.debug(f"Notes | Loading from: {path}")
        with open(path, "r") as f:
            notes = f.read()

        # Remove frontmatter
        notes = notes.split("---\n")
        notes = "---\n".join(notes[2:])

        return notes

    else:
        logger.debug(f"Notes | Creating new at: {path}")
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch(exist_ok=True)

        notes_to_write = (
            FRONTMATTER.substitute(
                timestamp=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                filename=path.stem,
            )
            + "\n"
            + DEFAULT_NOTES
        )

        # Save to file
        logger.debug(f"Notes | NEW | Saving to: {path}")
        with open(path, "w") as f:
            f.write(notes_to_write)

        return DEFAULT_NOTES


def notes_viewer() -> html.Div:
    """
    Notes viewer

    Returns:
        dash.html.Div: Div element containing a textarea for notes and a collapsible markdown viewer

    """
    return html.Div(
        [
            html.Div(
                [
                    html.H5(html.B("Notes"), className="is-size-5 is-4 is-pulled-left"),
                    # Toggle button to show/hide notes
                    html.Button(
                        html.I(className="fa-solid fa-eye-slash"),
                        id="hide-notes",
                        className="button is-info is-pulled-right",
                    ),
                ],
                className="column is-full mt-0",
                style={
                    "minHeight": "4rem",
                },
            ),
            html.Div(
                [
                    html.Button(
                        html.I(className="fa-solid fa-eye-slash"),
                        id="hide-preview",
                        className="button is-info is-pulled-left mx-2 my-0",
                        style={
                            "maxHeight": "280px",
                        },
                    ),
                    dcc.Textarea(
                        id="notes-area",
                        # placeholder="Type your notes here...",
                        value=DEFAULT_NOTES,
                        className="column ml-0 mr-2 p-3",
                        style={
                            "minHeight": "175px",
                            "maxHeight": "278px",
                        },
                    ),
                    html.Div(
                        dcc.Markdown(
                            id="notes-preview",
                            className="content m-0 p-0",
                            mathjax=True,
                            # NOTE: Local
                            dangerously_allow_html=True,
                            style={
                                "minHeight": "175px",
                                "maxHeight": "270px",
                                "overflowY": "scroll",
                            },
                        ),
                        id="notes-preview-div",
                        className="column m-2 p-3",
                        style={
                            "minHeight": "175px",
                            "maxHeight": "278px",
                            "overflowY": "scroll",
                        },
                    ),
                ],
                className="columns is-multiline box m-0 p-1",
                style={"maxHeight": "300px"},
                id="notes-viewer",
            ),
        ],
    )


@callback(
    Output("notes-preview-div", "style"),
    Output("hide-preview", "children"),
    Input("hide-preview", "n_clicks"),
)
def show_hide_preview(n_clicks: int) -> tuple:
    """
    Callback to show or hide the markdown preview

    Args:
        n_clicks (int): Number of times the button has been clicked

    Returns:
        tuple: Style to display the markdown preview and the button icon

    """
    if n_clicks is None or n_clicks % 2 == 0:
        return {"display": "block"}, html.I(className="fa-solid fa-eye-slash")
    else:
        return {"display": "none"}, html.I(className="fa-solid fa-eye")


@callback(
    Output("notes-preview", "children"),
    State("session-id", "data"),
    Input("notes-area", "value"),
)
def update_output(sess_id: str, new_notes: str) -> str:
    """
    Callback to update the markdown preview

    Args:
        sess_id (str): Session ID to load the state for
        new_notes (str): Value of the notes textarea

    Returns:
        str: Markdown preview to be displayed

    """
    _state = load_state_from_disk(sess_id)

    if new_notes is None:
        logger.debug("Notes | UPDATE | No notes to update")
        raise PreventUpdate
    # Save to file and update Last Saved
    base = Path(_state.measurement_path)
    if not _state.measurement_last_fmt:
        logger.debug("Notes | UPDATE | No measurement found")
        raise PreventUpdate

    path = base.parent / base.stem / (base.stem + ".md")

    if path.exists():
        logger.debug(f"Notes | UPDATE | Updating notes for: {path}")
        logger.debug(f"Notes | UPDATE | Updating notes with: {new_notes}")
        notes = (
            FRONTMATTER.substitute(
                timestamp=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                filename=path.stem,
            )
            + "\n"
            + new_notes
        )

        # Save to file
        logger.debug(f"Notes | UPDATE | Saving to: {path}")
        with open(path, "w") as f:
            f.write(notes)

        # Update preview
        return rf"{new_notes}"

    else:
        logger.debug(f"Notes | UPDATE | No notes found at: {path}")
        return DEFAULT_NOTES


@callback(
    # TODOLATER: Figure out how to push plots to the notes
    Output("notes-area", "value"),  # , allow_duplicate=True),
    State("session-id", "data"),
    Input("load-signal", "data"),
    # TODOLATER:
    # BUG: True required for `allow_duplicate`, but it makes the notes non-persistent.
    # BUG: But it is needed to allow multiple callbacks with `notes-area` as output. So, ¯\_(ツ)_/¯
    # prevent_initial_call=True,
)
def update_notes(sess_id: str, sig: int) -> str:
    """
    Callback to update notes & preview

    Args:
        sess_id (str): Session ID to load the state for
        sig (int): Signal to indicate that data has been updated

    Returns:
        str: Notes to be displayed

    """
    notes = _get_notes(sess_id)
    if sig in [None, 0]:
        return notes

    logger.debug("Notes | Updating notes as load-signal has changed.")
    return notes


@callback(
    Output("notes-viewer", "className", allow_duplicate=True),
    Output("hide-notes", "children"),
    Input("hide-notes", "n_clicks"),
    prevent_initial_call=True,
)
def show_hide_notes(n_clicks: int) -> tuple:
    """
    Callback to show or hide the notes viewer

    Args:
        n_clicks (int): Number of times the button has been clicked

    Returns:
        tuple: className to display the notes viewer and the updated button icon

    """
    if n_clicks is None or n_clicks % 2 == 0:
        return "columns is-multiline box m-0 p-1", html.I(
            className="fa-solid fa-eye-slash"
        )
    else:
        return "columns is-multiline box m-0 p-1 is-hidden", html.I(
            className="fa-solid fa-eye"
        )


# TODOLATER: Figure out how to push plots to the notes. See update_notes() above.
# @callback(
#     Output({"index": ALL, "type": "notes-store"}, "data", allow_duplicate=True),
#     Output("notes-area", "value", allow_duplicate=True),
#     Input({"index": ALL, "type": "notes-store"}, "data"),
#     prevent_initial_call=True,
# )
# def update_notes_from_store(notes_from_stores: list) -> Tuple[list, str]:
#     """
#     Callback to update notes from stores and to empty the stores after

#     Args:
#         notes_from_stores (list): List of notes from stores

#     Returns:
#         Tuple[list, str]: List of empty notes stores and the notes to be displayed

#     """
#     logger.debug(f"notes_from_stores : {notes_from_stores}")

#     # If notes found in stores, return the first one and empty all stores
#     empty_notes_stores = [None] * len(notes_from_stores)
#     for notes in notes_from_stores:
#         if notes:
#             logger.debug(f"Notes | Updating notes from store: {notes}")
#             return empty_notes_stores, notes

#     # If no notes found in stores, prevent update
#     logger.debug("Notes | No notes found in stores")
#     return empty_notes_stores, _get_notes()



import numpy as np
from copy import deepcopy
from numpy.polynomial.polynomial import Polynomial
from skimage import exposure
from scipy.signal import savgol_filter
from scipy.ndimage import affine_transform
from plotly import graph_objects as go

# Local imports
from qimchi.logger import logger
from qimchi.state import (
    DEFAULT_SAVGOL_OPTS,
    DEFAULT_SMA_WINDOW,
    DEFAULT_NORM_AXIS,
    DEFAULT_GC_OPTS,
    DEFAULT_LC_OPTS,
    DEFAULT_SC_OPTS,
    DEFAULT_POLYFIT_OPTS,
    DEFAULT_RI_OPTS,
    DEFAULT_ROTA_OPTS,
)
from qimchi.components.utils import format_number


def _safe_init(options: dict, key: str, default: dict) -> dict:
    """
    Helper function to safely initialize the options.

    Args:
        options (dict): Options dict.
        key (str): Key to get the value from the options dict.
        default (dict): Default options dict.

    Returns:
        dict: Safe options dict.

    """
    return options.get(key, default) if options else default


def _pprint_poly(poly: Polynomial, x_var: str):
    """
    Helper function to pretty-print a polynomial equation in LaTeX.

    Args:
        poly (Polynomial): Polynomial object.
        x_var (str): Variable name for the polynomial.

    Returns:
        str: Pretty-printed polynomial equation in LaTeX

    """

    # If x_var is too long, only use the capitalized letters in the string + the unit in parentheses
    if len(x_var) > 4:
        unit = x_var[x_var.find("(") :]
        x_var = x_var.split("(")[0].strip()
        x_var = "".join([c for c in x_var if c.isupper()])
        x_var += f"({unit})" if unit else ""

    # TODOLATER: Add support for auto-wrapped text
    for i, coef in enumerate(poly.coef):
        # # Format the coefficient for LaTeX, converting scientific notation if needed
        # if "e" in format_number(coef):
        #     base, exponent = format_number(abs(coef) if i > 0 else coef).split("e")
        #     formatted_coef = f"{base}\\times 10^{{{int(exponent)}}}"
        # else:
        #     formatted_coef = format_number(abs(coef) if i > 0 else coef)

        if i == 0:
            poly_str = f"{format_number(coef)}"
            # poly_str = f"{formatted_coef}"

        else:
            sign = "+" if coef >= 0 else "-"
            poly_str += f" {sign} {format_number(abs(coef))}({x_var})^{i}"

    logger.debug(f"Pretty-printed polynomial equation: {poly_str}")
    return f"{poly_str}"


class Filter:
    def __init__(
        self,
        figure: go.Figure,
        num_axes: int,
        options: dict = None,
    ) -> None:
        """
        Filter Base Class

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure
            options (dict, optional): Options for the filter. Default is None.

        """
        self._name = self.__class__.__name__

        self.figure = figure
        self.new_fig = deepcopy(figure)
        self.data = figure["data"]
        self.num_axes = num_axes
        self.options = options

        dat = figure["data"][0]
        self.x_axis = np.array(dat["x"])
        self.y_axis = np.array(dat["y"])
        self.x_label = figure["layout"]["xaxis"]["title"]["text"]
        self.y_label = figure["layout"]["yaxis"]["title"]["text"]
        self.title = figure["layout"]["title"]["text"]
        if self.num_axes == 2:
            self.z_axis = np.array(dat["z"])
            self.z_label = figure["layout"]["coloraxis"]["colorbar"]["title"]["text"]

    def apply(self, *args, **kwargs):
        """
        Calls the appropriate apply method based on the number of axes.

        Raises:
            NotImplementedError: If the number of axes is not 1 or 2.

        """
        try:
            match self.num_axes:
                case 1:
                    self.apply_1d(*args, **kwargs)
                case 2:
                    self.apply_2d(*args, **kwargs)
                case _:
                    logger.error(
                        f"Filtering not supported for {self.num_axes}D plots. Only 1D and 2D plots are supported."
                    )
                    raise NotImplementedError(
                        f"Filtering not supported for {self.num_axes}D plots. Only 1D and 2D plots are supported."
                        # Re-scale the colorbar
                    )
        except (NotImplementedError, Exception) as err:
            # TODOLATER: Connect to Toast notifications
            logger.error(err, exc_info=True)

        return self.new_fig

    def _update_title(self, fil: str) -> None:
        """
        Updates the plot title with the filter name.

        Args:
            fil (str): Filter name to append to plot title.

        """
        ttl = self.title
        title_default = " ".join(ttl.split(" ")[:4])
        tfl: str = fil
        if "Filt." in ttl:
            # NOTE: plot_title - Filt.: fil1 fil2 ...
            # Default title - f"{_state.device_type} {_state.wafer_id} {_state.device_id} {_state.measurement}",
            # [4] is "Fil:"
            tfl: list[str] = ttl.split(" ")[6:]
            tfl.insert(0, fil)
            tfl: str = " ".join(tfl)

        # Update the title
        self.new_fig["layout"]["title"]["text"] = f"{title_default} <br> Filt.: {tfl}"

    def _hmap_update(self, z_data: np.ndarray, fil: str) -> None:
        """
        Updates the 2D HeatMap plot with new Z-axis data & label and re-scales the colorbar.

        Args:
            z_data (np.ndarray): Z-axis data.
            fil (str): Filter name to append to plot title.

        Returns:
            go.Figure: Updated figure.

        """
        fig_caxis = self.new_fig["layout"]["coloraxis"]
        fig_caxis["colorbar"]["title"]["text"] = (
            f"Filt.<br>{self.z_label}" if "Filt." not in self.z_label else self.z_label
        )

        fig_caxis["cmin"] = np.nanmin(z_data)
        fig_caxis["cmax"] = np.nanmax(z_data)

        # Update the figure
        self.new_fig["data"][0]["z"] = z_data
        self.new_fig["layout"]["coloraxis"] = fig_caxis
        self._update_title(fil)


class FlipHeatMap(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Flip HeatMap Filter by multiplying the Z-axis data by -1.
        This is useful for inverting the color scale of a heatmap.

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)

    def apply_1d(self):
        """
        Not supported for 1D plots.

        """
        raise NotImplementedError("Flip  not supported for 1D plots.")

    def apply_2d(self):
        """
        Applies the flip heatmap filter to a 2D plot.

        """
        self._hmap_update(-self.z_axis, fil="Flip")


class Differentiate(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Differentiation Filter

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)

    def apply_1d(self, *_):
        """
        Applies the differentiation filter to a 1D plot.

        """
        self.new_fig["data"][0]["y"] = np.gradient(self.y_axis, self.x_axis)
        self.new_fig["layout"]["yaxis"]["title"]["text"] = f"d{self.y_label}"
        self._update_title("d")

    def apply_2d(self, twod_axis=0):
        """
        Apples the differentiation filter to a 2D plot.

        Args:
            twod_axis (int): Axis along which to differentiate

        """
        z_data = np.gradient(self.z_axis, self.x_axis, self.y_axis)[twod_axis]

        match twod_axis:
            case 0:
                twod_axis_label = (
                    f"d{self.x_label}" if "$" not in self.x_label else "dx"
                )
            case 1:
                twod_axis_label = (
                    f"d{self.y_label}" if "$" not in self.y_label else "dy"
                )
            case _:
                err = f"Invalid value of `twod_axis={twod_axis}` for differentiation."
                logger.error(err)
                raise ValueError(err)

        # Update the figure
        self._hmap_update(z_data, fil=twod_axis_label)


class Smooth(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Smoothing Filter

        Features:
            - Savitzky-Golay filter

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)
        self.window = _safe_init(options, "window", DEFAULT_SAVGOL_OPTS["window"])
        self.polyorder = _safe_init(
            options, "polyorder", DEFAULT_SAVGOL_OPTS["polyorder"]
        )
        self.smooth_axis = _safe_init(options, "axis", DEFAULT_SAVGOL_OPTS["axis"])
        self.deriv = _safe_init(options, "deriv", DEFAULT_SAVGOL_OPTS["deriv"])
        self.mode = _safe_init(options, "mode", DEFAULT_SAVGOL_OPTS["mode"])
        self.cval = _safe_init(options, "cval", DEFAULT_SAVGOL_OPTS["cval"])
        self.delta = _safe_init(options, "delta", DEFAULT_SAVGOL_OPTS["delta"])

    def apply_1d(self):
        """
        Applies the smoothing filter to a 1D plot.

        """
        self.new_fig["data"][0]["y"] = savgol_filter(
            self.y_axis,
            window_length=self.window,
            polyorder=self.polyorder,
        )
        self.new_fig["layout"]["yaxis"]["title"]["text"] = f"Sm {self.y_label}"
        self._update_title("Sm")

    def apply_2d(self):
        """
        Applies the smoothing filter to a 2D plot.

        """
        logger.debug(
            f"Applying Smooth Filter (SavGol) with window={self.window}, polyorder={self.polyorder}, axis={self.smooth_axis}..."
        )
        if self.smooth_axis == 2:
            # Smooth along the z-axis (depth) by applying Savitzky-Golay filter twice
            # First, smooth along the x-axis (rows) and then along the y-axis (columns)
            z_data_x = savgol_filter(
                self.z_axis,
                window_length=self.window,
                polyorder=self.polyorder,
                axis=0,
                deriv=self.deriv,
                delta=self.delta,
                mode=self.mode,
                cval=self.cval,
            )
            z_data = savgol_filter(
                z_data_x,
                window_length=self.window,
                polyorder=self.polyorder,
                axis=1,
                deriv=self.deriv,
                delta=self.delta,
                mode=self.mode,
                cval=self.cval,
            )
        else:
            z_data = savgol_filter(
                self.z_axis,
                window_length=self.window,
                polyorder=self.polyorder,
                axis=self.smooth_axis,
                deriv=self.deriv,
                delta=self.delta,
                mode=self.mode,
                cval=self.cval,
            )

        # Update the figure
        self._hmap_update(z_data, fil="Sm")


class SimpleMovingAverage(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Simple Moving Average Filter.
        NOTE: Trailing: Output length is `window - 1` less than the input length.

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)
        self.window = _safe_init(options, "window", DEFAULT_SMA_WINDOW)

    def apply_1d(self):
        """
        Applies the simple moving average filter to a 1D plot.

        """
        # NOTE: Trailing: Output length is `window - 1` less than the input length.
        self.new_fig["data"][0]["y"] = np.convolve(
            self.y_axis, np.ones(self.window) / self.window, mode="valid"
        )
        self.new_fig["layout"]["yaxis"]["title"]["text"] = f"SMA {self.y_label}"
        self._update_title("SMA")

    def apply_2d(self):
        """
        Not supported for 2D plots.

        """
        raise NotImplementedError("Simple Moving Average not supported for 2D plots.")


class Normalize(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Normalization Filter

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)
        # str: "x" or "y" or "z"
        self.axis = _safe_init(options, "axis", DEFAULT_NORM_AXIS)

    def apply_1d(self):
        """
        Applies the normalization filter to a 1D plot.

        """
        self.new_fig["data"][0]["y"] = self.y_axis / np.nanmax(self.y_axis)
        self.new_fig["layout"]["yaxis"]["title"]["text"] = f"Norm {self.y_label}"
        self._update_title("Norm")

    def apply_2d(self):
        """
        Applies the normalization filter to a 2D plot.

        """
        logger.debug(f"Normalizing 2D plot along axis `{self.axis}`.")

        match self.axis:
            case "z":
                z_data = self.z_axis / np.nanmax(self.z_axis)
            # NOTE: [i, j] = [row, col] = [y, x]
            case "x":
                # Normalize each column (along x-axis)
                z_data = self.z_axis / np.nanmax(self.z_axis, axis=0)
            case "y":
                # Normalize each row (along y-axis)
                z_data = self.z_axis / np.nanmax(self.z_axis, axis=1)[..., np.newaxis]
            case _:
                err = f"Invalid value of `axis={self.axis}` for normalization."
                logger.error(err)

        # Update the figure
        self._hmap_update(z_data, fil=f"Norm({self.axis})")


class GammaCorrection(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Gamma Correction Filter. Uses `skimage.exposure.adjust_gamma()`.

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)
        logger.debug(f"GammaCorrection | options: {options}")
        self.gamma = _safe_init(options, "gamma", DEFAULT_GC_OPTS["gamma"])
        self.gain = _safe_init(options, "gain", DEFAULT_GC_OPTS["gain"])

    def apply_1d(self):
        """
        Not supported for 1D plots.

        """
        raise NotImplementedError("Gamma Correction not supported for 1D plots.")

    def apply_2d(self):
        """
        Applies the gamma correction filter to a 2D plot.

        """
        logger.debug(
            f"Applying Gamma Correction with gamma={self.gamma}, gain={self.gain}..."
        )
        z_data = exposure.adjust_gamma(self.z_axis, self.gamma, self.gain)

        # Update the figure
        self._hmap_update(z_data, fil="γC")


class LogCorrection(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Logarithmic Correction Filter. Uses `skimage.exposure.adjust_log()`.

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)
        self.gain = _safe_init(options, "gain", DEFAULT_LC_OPTS["gain"])
        self.inv = _safe_init(options, "inv", DEFAULT_LC_OPTS["inv"])

    def apply_1d(self):
        """
        Not supported for 1D plots.

        """
        raise NotImplementedError("Logarithmic Correction not supported for 1D plots.")

    def apply_2d(self):
        """
        Applies the logarithmic correction filter to a 2D plot.

        """
        logger.debug(
            f"Applying Logarithmic Correction with gain={self.gain}, inv={self.inv}..."
        )
        z_data = exposure.adjust_log(self.z_axis, gain=self.gain, inv=self.inv)

        # Update the figure
        self._hmap_update(z_data, fil="logC")


class SigmoidCorrection(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Sigmoid Correction Filter. Uses `skimage.exposure.adjust_sigmoid()`.

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)
        self.cutoff = _safe_init(options, "cutoff", DEFAULT_SC_OPTS["cutoff"])
        self.gain = _safe_init(options, "gain", DEFAULT_SC_OPTS["gain"])

    def apply_1d(self):
        """
        Not supported for 1D plots.

        """
        raise NotImplementedError("Sigmoid Correction not supported for 1D plots.")

    def apply_2d(self):
        """
        Applies the sigmoid correction filter to a 2D plot.

        """
        logger.debug(
            f"Applying Sigmoid Correction with cutoff={self.cutoff}, gain={self.gain}..."
        )
        z_data = exposure.adjust_sigmoid(
            self.z_axis, cutoff=self.cutoff, gain=self.gain
        )
        # Update the figure
        self._hmap_update(z_data, fil="σC")


class RescaleIntensity(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Rescale Intensity Filter. Uses `skimage.exposure.rescale_intensity()`.

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)
        self.in_range = _safe_init(options, "in_range", DEFAULT_RI_OPTS["in_range"])

    def apply_1d(self):
        """
        Not supported for 1D plots.

        """
        raise NotImplementedError("Rescale Intensity not supported for 1D plots.")

    def apply_2d(self):
        """
        Applies the rescale intensity filter to a 2D plot. Uses the min-max scaling method.
        # TODOLATER: Can allow specifying the range to scale to. See: https://scikit-image.org/docs/stable/api/skimage.exposure.html#skimage.exposure.rescale_intensity

        """
        logger.debug("Applying Rescale Intensity...")
        z_data = exposure.rescale_intensity(self.z_axis)

        # Update the figure
        self._hmap_update(z_data, fil="ReC")


class LogScale(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Logarithmic Scaling Filter.

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)

    def apply_1d(self):
        """
        Applies the logarithmic scaling filter to a 1D plot.

        """
        self.new_fig["data"][0]["y"] = np.log(self.y_axis)
        self.new_fig["layout"]["yaxis"]["title"]["text"] = f"log {self.y_label}"
        self._update_title("log")

    def apply_2d(self):
        """
        Applies the logarithmic scaling filter to a 2D plot.

        """
        self._hmap_update(np.log(self.z_axis), fil="log")


class PolyFit(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Polynomial Fitting Filter. Uses `numpy.polynomial.polynomial.Polynomial.fit()`.

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)
        self.deg = _safe_init(options, "deg", DEFAULT_POLYFIT_OPTS["deg"])
        self.window = _safe_init(options, "window", DEFAULT_POLYFIT_OPTS["window"])

    def apply_1d(self):
        """
        Applies the polynomial fitting filter to a 1D plot.

        """
        # Fit the data to a polynomial
        poly = Polynomial.fit(
            self.x_axis,
            self.y_axis,
            deg=self.deg,
            window=self.window,
        )
        fit_line = poly(self.x_axis)

        # Add the fit line to the plot
        self.new_fig.add_trace(
            go.Scatter(
                x=self.x_axis,
                y=fit_line,
                mode="lines",
                # TODOLATER: @s.anupam Standardize the color scheme & name etc.
                line=dict(color="rebeccapurple", width=2),
                # TODOLATER: Add to legend if needed. CUrrently hidden.
                showlegend=False,
                name=f"Fit (deg={self.deg})",
            )
        )

        # Add the equation to the plot
        # TODOLATER: Add a helper function that does the following::
        # TODOLATER: - [x] Add pretty-printed (LaTeX) polynomial equation with wrapped text
        # TODOLATER: - [ ] Add support for auto-wrapped text - See _pprint_poly()
        self.new_fig.add_annotation(
            # Center
            x=0.95,
            y=0.95,
            xref="paper",
            yref="paper",
            text=f"{_pprint_poly(poly, self.x_label)}",
            showarrow=False,
            font=dict(size=16, color="rebeccapurple"),
            bgcolor="white",
            # TODOLATER: Add if needed
            # BUG: Displaces the LaTeX-ified text
            # bordercolor="rebeccapurple",
            # borderwidth=1,
            # borderpad=4,
            # opacity=0.8,
            # width=200,  # To fit the text (text can overflow and get cut off)
        )

    def apply_2d(self):
        """
        Not supported for 2D plots.

        """
        raise NotImplementedError("Polynomial Fitting not supported for 2D plots.")


class RotateHeatMap(Filter):
    def __init__(self, figure: go.Figure, num_axes: int, options: dict = None) -> None:
        """
        Rotates HeatMap by specified angle.

        Args:
            figure (go.Figure): Plotly Figure object to apply filters to.
            num_axes (int): Number of axes in the figure.
            options (dict, optional): Options for the filter. Default is None.

        """
        super().__init__(figure, num_axes, options)
        self.angle = _safe_init(options, "angle", DEFAULT_ROTA_OPTS["angle"])

    def apply_1d(self):
        """
        Not supported for 1D plots.

        """
        raise NotImplementedError("Rotate HeatMap not supported for 1D plots.")

    def apply_2d(self):
        """
        Applies the rotate heatmap filter to a 2D plot.

        """
        logger.debug(f"Rotating HeatMap by {self.angle} degrees...")

        # Translate the data to the origin and then rotate
        z_data = self.z_axis

        # Rotation matrix
        angle_rad = np.deg2rad(self.angle)
        c, s = np.cos(angle_rad), np.sin(angle_rad)
        rota_mat = np.array([[c, s], [-s, c]])

        # New bounding box after rotation to prevent cropping
        # Thanks: rotate():https://github.com/scipy/scipy/blob/v1.14.1/scipy/ndimage/_interpolation.py#L865-L1001
        img_shape = np.asarray(z_data.shape)
        iy, ix = img_shape
        out_bounds = rota_mat @ [[0, 0, iy, iy], [0, ix, 0, ix]]

        # Shape of the transformed input plane
        out_plane_shape = (np.ptp(out_bounds, axis=1) + 0.5).astype(int)

        # Final output center after rotation
        out_center = rota_mat @ ((out_plane_shape - 1) / 2)
        in_center = (img_shape - 1) / 2
        final_offset = in_center - out_center

        # Use affine_transform to apply the rotation with the calculated offset and output shape
        z_rotated = affine_transform(
            z_data,
            rota_mat,
            offset=final_offset,
            output_shape=tuple(out_plane_shape),
            order=0,  # Nearest neighbor interpolation
            mode="constant",
            cval=np.nan,  # Empty space with np.nan (white)
            prefilter=False,
        )
        # Update the axis labels
        self.new_fig["layout"]["xaxis"]["title"]["text"] = (
            rf"${c:.3f}\text{{{self.x_label}}} + {-s:.3f}\text{{{self.y_label}}}\:({self.angle}^{{\circ}})$"
        )
        self.new_fig["layout"]["yaxis"]["title"]["text"] = (
            rf"${s:.3f}\text{{{self.x_label}}} + {c:.3f}\text{{{self.y_label}}}\:({self.angle}^{{\circ}})$"
        )

        # Update the figure
        self._hmap_update(z_rotated, fil=f"Rot({self.angle}°)")


def apply_filters(
    filters_order: list, filters_opts: dict, fig: go.Figure, fig_num_axes: int
) -> go.Figure:
    """
    Applies the filters to the figure.

    Args:
        filters_order (list): List of filters to apply, in order.
        filters_opts (dict): Dict of filters to apply, with corresponding options.
        fig (go.Figure): Plotly Figure object to apply filters to.
        fig_num_axes (int): Number of axes in the figure.

    Returns:
        go.Figure: Filtered figure.

    """
    filt_fig = None
    fig_tmp = deepcopy(fig)
    for fil in filters_order:
        # `filter_opts` is a nested dict. Get the required dict and then pass it
        opts = filters_opts[fil]
        logger.debug(f"apply_filters | Filter: {fil} | Options: {opts}")
        match fil:
            case "flip":
                logger.debug("apply_filters | Applying `Flip` filter...")
                filt_obj = FlipHeatMap(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "diff":
                logger.debug("apply_filters | Applying `Differentiate` (1D) filter...")
                filt_obj = Differentiate(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()
            case "diff_x":
                logger.debug(
                    "apply_filters | Applying `Differentiate` (2D - x) filter..."
                )
                filt_obj = Differentiate(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply(twod_axis=0)
            case "diff_y":
                logger.debug(
                    "apply_filters | Applying `Differentiate` (2D - y) filter..."
                )
                filt_obj = Differentiate(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply(twod_axis=1)

            case "savgol":
                logger.debug("apply_filters | Applying `Smooth` filter...")
                filt_obj = Smooth(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "sma":
                logger.debug(
                    "apply_filters | Applying `Simple Moving Average` filter..."
                )
                filt_obj = SimpleMovingAverage(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "normalize":
                logger.debug("apply_filters | Applying `Normalize` filter...")
                filt_obj = Normalize(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "gamma_corr":
                logger.debug("apply_filters | Applying `Gamma Correction` filter...")
                filt_obj = GammaCorrection(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "log_corr":
                logger.debug(
                    "apply_filters | Applying `Logarithmic Correction` filter..."
                )
                filt_obj = LogCorrection(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "sig_corr":
                logger.debug("apply_filters | Applying `Sigmoid Correction` filter...")
                filt_obj = SigmoidCorrection(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "rescale_intensity":
                logger.debug("apply_filters | Applying `Rescale Intensity` filter...")
                filt_obj = RescaleIntensity(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "log_scale":
                logger.debug("apply_filters | Applying `Logarithmic Scaling` filter...")
                filt_obj = LogScale(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "rotate":
                logger.debug("apply_filters | Applying `Rotate HeatMap` filter...")
                filt_obj = RotateHeatMap(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case "polyfit":
                logger.debug("apply_filters | Applying `Polynomial Fitting` filter...")
                filt_obj = PolyFit(fig_tmp, fig_num_axes, opts)
                filt_fig = filt_obj.apply()

            case _:
                err = f"No definition for Filter {fil} was found."
                logger.error(err)
                raise ValueError(err)
        fig_tmp = filt_fig

    if filt_fig is None and not filters_order:
        # If no filters were applied, return the original figure
        logger.debug("No filters were applied.")
        return fig
    return filt_fig



from pathlib import Path
from abc import ABC, abstractmethod
from dash import html


class Database(ABC):
    def __init__(self, path: Path) -> None:
        """
        Base class for database folders in Qimchi

        """
        super().__init__()
        self.path = path

    @abstractmethod
    def options(self):
        """
        Returns the options for the data selector for current database type

        """

    def selector(self) -> html.Div:
        """
        Returns the data selector component

        Returns:
            html.Div: The data selector component

        """
        return html.Div(
            [
                self.options(),
            ],
            **{"data-theme": "light"},
        )


class Data(ABC):
    def __init__(self, path: Path) -> None:
        """
        Base class for data files in Qimchi

        """
        super().__init__()
        self.path = path



from pathlib import Path
from typing import Any, List

import dash
from dash import Input, Output, State, callback, dcc, html
from dash.exceptions import PreventUpdate

# Local imports
from ..base import Data, Database
from qimchi.state import load_state_from_disk
from qimchi.logger import logger
from qimchi.components.utils import read_data


def _update_options_fn(
    src: str,
    condition_list: List[str | Path | Any],
    check_file: bool = False,
    default_return: List[str] | None = [],
) -> List[str] | str | None:
    """
    Helper function to update options based on the condition list.

    Args:
        src: Source of the update.
        condition_list (List[str | Path | Any]): List of parameters to compare with None.
            Ordering matters, as the path is constructed from the first item to the second last item.
            The last item is always assumed to be `dataset_type`
        check_file (bool, optional): Check if the path is a file, by default False (checks for directory)
        default_return (List | None, optional): Default return item if the condition is not satisfied, by default []

    Returns:
        List[str] | str | None: List of strings (options) or empty list or None

    """
    logger.debug(f"_update_options_fn | src: {src}")

    # NOTE: This only checks for truthy (i.e., "" is False, but Path("") is True). Should work here though.
    if all(condition_list):
        # Assuming either strings or Paths
        path = Path(*condition_list[:-1])

        if not path.is_dir():
            logger.debug(f"Path {path} is not a directory.")
            return default_return
        # Unfiltered options
        options: List[str] = [p.name for p in path.iterdir()]
        # Filter for only files or directories
        options = [
            opt
            for opt in options
            if (
                # TODOLATER: Some redundancy here
                (path / opt).is_dir() and (path / opt).suffix == ".zarr"
                if check_file
                else (path / opt).is_dir()
            )
        ]
        logger.debug(f"_update_options_fn | src: {src} | options: {options}")
        return options

    logger.debug(f"_update_options_fn | src: {src} | default_return: {default_return}")
    return default_return


def _should_update_data(sess_id: str, path: Path, sig_curr: int, origin: str) -> bool:
    """
    Conditionally update the data store with the Zarr file content.

    Args:
        sess_id (str): Session ID to load the state for
        path (Path): Path to the dataset
        sig_curr (int): The current signal
        origin (str): Origin of the data update

    Returns:
        bool: Whether the data has been updated

    """
    _state = load_state_from_disk(sess_id)

    if path.is_dir() and path.suffix == ".zarr":
        curr_mod_time = path.stat().st_mtime
        if str(path) == str(_state.measurement_path):
            if curr_mod_time != _state.measurement_last_fmt:
                logger.warning(
                    f"_should_update_data | {sess_id} | {origin}: DATA UPDATED ON DISK!! AT {path}."
                )
                _state.measurement_last_fmt = curr_mod_time
                _state.save_state()
                return True

            # If the data has not been loaded even once, load it (e.g., for the first time on page load)
            if sig_curr in [None, 0]:
                logger.warning(
                    f"_should_update_data | {sess_id} | {origin}: DATA BEING LOADED FROM {path}"
                )
                return True
            # If the data has been loaded, do not update
            else:
                return False

        else:
            # If the path has changed, update the state
            logger.warning(
                f"_should_update_data | {sess_id} | {origin}: DATA PATH MODIFIED!! AT {path}."
            )
            _state.measurement_path = path
            _state.measurement_last_fmt = curr_mod_time
            _state.save_state()

            if origin == "XarrayData":
                # To update the wafer_id etc. in the state
                data = read_data(sess_id, src="_should_update_data")
                metadata = data.attrs
                device_type = metadata.get("Device Type", "")
                wafer_id = metadata.get("Wafer ID", "")
                sample_name = metadata.get("Sample Name", "")
                meas_id = metadata.get("Measurement ID", "")

                # Update the state
                _state.device_type = device_type
                _state.wafer_id = wafer_id
                _state.device_id = sample_name
                _state.measurement = meas_id
                _state.save_state()

            logger.warning(
                f"_should_update_data | {sess_id} | {origin}: DATA LOADED FROM MODIFIED {path}"
            )
            return True


class XarrayDataFolder(Database):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @staticmethod
    @callback(
        State("session-id", "data"),
        Input("measurement", "value"),
        Input("measurement-type", "value"),
        Input("device-id", "value"),
        Input("device-type", "value"),
        Input("wafer-id", "value"),
    )
    def update_state_dropdown_vals(
        sess_id: str,
        measurement: str,
        measurement_type: str,
        device_id: str,
        device_type: str,
        wafer_id: str,
    ) -> None:
        """
        Update the state with the wafer id, device type, device id, measurement type, and measurement.

        """
        _state = load_state_from_disk(sess_id)

        _state.wafer_id = wafer_id if wafer_id else _state.wafer_id
        _state.device_type = device_type if device_type else _state.device_type
        _state.device_id = device_id if device_id else _state.device_id
        _state.measurement_type = (
            measurement_type if measurement_type else _state.measurement_type
        )
        _state.measurement = measurement if measurement else _state.measurement
        _state.save_state()
        logger.debug(
            f"update_state_dropdown_vals | Saved state for {sess_id=}"  # | {_state=}"
        )

    @staticmethod
    @callback(
        Output("is-data-selector-set", "data"),
        Output("wafer-id", "options"),
        State("wafer-id", "options"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        prevent_initial_call=True,
    )
    def update_wafer_id(
        wafer_options: List[str],
        dataset_path: str | Path,
        dataset_type: str,
        *_,
    ) -> List[str]:
        """
        Update the wafer ID options based on the dataset path and type.
        This function is triggered by the submit button and updates the wafer ID options in the dropdown.
        This in turn triggers other dropdowns sequentially.

        Args:
            wafer_options (List[str]): Current wafer ID options
            dataset_path (str | Path): Path to the dataset
            dataset_type (str): Type of the dataset

        Returns:
            bool: Whether the data selector is set
            List[str]: Updated wafer ID

        """
        input_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        logger.debug(f"update_wafer_id || Input ID: {input_id}")
        logger.debug(
            f"update_wafer_id | Dataset Path: {dataset_path} | Dataset Type: {dataset_type}"
        )
        logger.debug(f"update_wafer_id || Wafer Options: {wafer_options}")

        # If wafer options are already set, return them, if input is from submit
        if input_id == "submit" and wafer_options:
            logger.debug(
                f"update_wafer_id || Wafer options already set: {wafer_options} | Preventing update."
            )
            raise PreventUpdate

        key = "wafer_id"
        opts = _update_options_fn(
            key,
            [dataset_path, dataset_type],
            default_return=wafer_options,
        )

        return bool(opts), opts

    @staticmethod
    @callback(
        Output("device-type", "options"),
        State("device-type", "options"),
        Input("wafer-id", "value"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        prevent_initial_call=True,
    )
    def update_device_type(
        device_type_options: List[str],
        wafer_id: str,
        dataset_path: str | Path,
        dataset_type: str,
        *_,
    ) -> List[str] | None:
        input_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        logger.debug(f"update_device_type || Input ID: {input_id}")

        key = "device_type"

        return _update_options_fn(
            key,
            [dataset_path, wafer_id, dataset_type],
        )

    @staticmethod
    @callback(
        Output("device-id", "options"),
        State("device-id", "options"),
        Input("device-type", "value"),
        Input("wafer-id", "value"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        prevent_initial_call=True,
    )
    def update_device_id(
        device_id_options: List[str],
        device_type: str,
        wafer_id: str,
        dataset_path: str | Path,
        dataset_type: str,
        *_,
    ) -> List[str] | None:
        input_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        logger.debug(f"update_device_id || Input ID: {input_id}")

        key = "device_id"

        return _update_options_fn(
            key,
            [dataset_path, wafer_id, device_type, dataset_type],
        )

    @staticmethod
    @callback(
        Output("measurement-type", "options"),
        State("measurement-type", "options"),
        Input("device-id", "value"),
        Input("device-type", "value"),
        Input("wafer-id", "value"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        prevent_initial_call=True,
    )
    def update_measurement_type(
        measurement_type_options: List[str],
        device_id: str,
        device_type: str,
        wafer_id: str,
        dataset_path: str | Path,
        dataset_type: str,
        *_,
    ) -> List[str] | None:
        input_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        logger.debug(f"update_measurement_type || Input ID: {input_id}")

        key = "measurement_type"

        return _update_options_fn(
            key,
            [dataset_path, wafer_id, device_type, device_id, dataset_type],
        )

    @staticmethod
    @callback(
        Output("measurement", "options"),
        State("measurement", "options"),
        Input("measurement-type", "value"),
        Input("device-id", "value"),
        Input("device-type", "value"),
        Input("wafer-id", "value"),
        Input("dataset-path", "value"),
        Input("dataset-type", "value"),
        Input("submit", "n_clicks"),
        Input("upload-ticker", "n_intervals"),
        prevent_initial_call=True,
    )
    def update_measurement(
        measurement_options: List[str],
        measurement_type: str,
        device_id: str,
        device_type: str,
        wafer_id: str,
        dataset_path: str | Path,
        dataset_type: str,
        *_,
    ) -> List[str] | None:
        input_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        logger.debug(f"update_measurement || Input ID: {input_id}")

        key = "measurement"
        options = _update_options_fn(
            key,
            [
                dataset_path,
                wafer_id,
                device_type,
                device_id,
                measurement_type,
                dataset_type,
            ],
            check_file=True,  # Checks if the path is a measurement file
        )
        options.sort(key=lambda x: int(x.split("-")[0]))

        # If measurement_options is None, return new options
        if not measurement_options:
            return options

        # Else, if the input_id is "upload-ticker", check if the options are the same
        if input_id == "upload-ticker":
            # If same, PreventUpdate
            if set(measurement_options) == set(options):
                logger.debug(
                    "update_measurement | Measurement options are the same. Preventing update."
                )
                raise PreventUpdate
            # If not the same, return the updated options
            else:
                logger.debug(
                    "update_measurement | Measurement options are different. Updating measurement options."
                )
                return options
        return options

    @staticmethod
    @callback(
        Output(
            "load-signal",
            "data",
            allow_duplicate=True,
        ),
        State("session-id", "data"),
        State("load-signal", "data"),
        Input(
            "measurement",
            "value",
        ),
        State("measurement-type", "value"),
        State("device-id", "value"),
        State("device-type", "value"),
        State("wafer-id", "value"),
        State("dataset-path", "value"),
        Input("upload-ticker", "n_intervals"),
        prevent_initial_call=True,
    )
    def update_data(
        sess_id: str,
        sig_curr: int,
        measurement: str,
        measurement_type: str,
        device_id: str,
        device_type: str,
        wafer_id: str,
        dataset_path: str | Path,
        *_,
    ) -> dict:
        """
        Update the data store with the Zarr file content.

        Args:
            sess_id (str): Session ID
            sig_curr (int): The current signal value
            measurement (str): Measurement name
            measurement_type (str): Measurement type
            device_id (str): Device ID
            device_type (str): Device type
            wafer_id (str): Wafer ID
            dataset_path (str | Path): Path to the dataset

        Returns:
            dict: The updated data store

        Raises:
            PreventUpdate: If the path is not a file or not a Zarr file

        """
        input_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        logger.debug(f"update_data | Input ID: {input_id}")
        logger.debug(
            f"update_data | Measurement: {measurement} | Measurement Type: {measurement_type} | Device ID: {device_id} | Device Type: {device_type} | Wafer ID: {wafer_id} | Dataset Path: {dataset_path}"
        )

        if all(
            [
                measurement,
                measurement_type,
                device_id,
                device_type,
                wafer_id,
                dataset_path,
            ]
        ):
            measurement_path = Path(
                dataset_path,
                wafer_id,
                device_type,
                device_id,
                measurement_type,
                measurement,
            )
            path = measurement_path
            if _should_update_data(sess_id, path, sig_curr, "XarrayDataFolder"):
                return sig_curr + 1 if sig_curr is not None else 1
            else:
                raise PreventUpdate

    @staticmethod
    @callback(
        Output("measurement", "value"),
        Input("prev-mmnt-button", "n_clicks"),
        Input("next-mmnt-button", "n_clicks"),
        Input("measurement", "value"),
        State("measurement", "options"),
        State("session-id", "data"),
    )
    def update_selected_measurement(
        _prev_clicks: int,
        _next_clicks: int,
        curr_mmnt: str,
        mmnt_options: list,
        sess_id: str,
    ) -> int:
        """
        Updates selected measurement based on button clicks or dropdown selection.

        Args:
            _prev_clicks (int): Number of clicks on the previous button
            _next_clicks (int): Number of clicks on the next button
            curr_mmnt (str): Current selected measurement
            mmnt_options (list): List of measurement options
            sess_id (str): Session ID


        Returns:
            str: New selected measurement

        Raises:
            PreventUpdate: If the options are not available

        """
        _state = load_state_from_disk(sess_id)

        input_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]
        logger.debug(f"update_selected_measurement | Input ID: {input_id}")
        logger.debug(
            f"update_selected_measurement | {dash.callback_context.triggered[0]=}"  # '.'
        )

        # BUG: Upstream bug. See https://github.com/plotly/dash/issues/1523
        if not input_id:
            logger.debug("update_selected_measurement | No button clicked")

            old_mmnt = _state.measurement
            # NOTE: Assuming this is None only on reload
            logger.debug(
                f"update_selected_measurement | No button clicked | Options not available: Returning old measurement: {old_mmnt}"
            )

            if old_mmnt:
                return old_mmnt
            else:
                raise PreventUpdate

        if not mmnt_options:
            old_mmnt = _state.measurement
            # NOTE: Assuming this is None only on reload
            logger.debug(
                f"Options not available: Returning old measurement: {old_mmnt}"
            )
            return old_mmnt

        # Identify which button was pressed
        triggered_id = dash.callback_context.triggered[0]["prop_id"].split(".")[0]

        # If no measurement is selected, do not update
        if not curr_mmnt:
            logger.debug(
                "update_selected_measurement | No current measurement selected"
            )
            raise PreventUpdate

        curr_idx = mmnt_options.index(curr_mmnt)
        if triggered_id == "prev-mmnt-button":
            new_index = (curr_idx - 1) % len(mmnt_options)  # Cycle left
        elif triggered_id == "next-mmnt-button":
            new_index = (curr_idx + 1) % len(mmnt_options)  # Cycle right
        else:
            new_index = curr_idx  # No change

        new_meas = mmnt_options[new_index]
        _state.measurement = new_meas
        _state.save_state()

        logger.debug(f"New index: {new_index}")
        logger.debug(f"New option: {new_meas}")

        return new_meas

    def options(self):
        """
        Dropdowns in the selector.

        Returns:
            dash.html.Div: Div element containing the dropdowns for the selector

        """
        return html.Div(
            [
                # Label
                html.Div(
                    "Sample Info:",
                    className="column is-1",
                ),
                # Dropdowns
                html.Div(
                    dcc.Dropdown(
                        id="wafer-id",
                        placeholder="Wafer ID",
                        searchable=True,
                        persistence=True,
                        persistence_type="session",
                    ),
                    className="column is-2",
                ),
                html.Div(
                    dcc.Dropdown(
                        id="device-type",
                        placeholder="Device Type",
                        searchable=True,
                        persistence=True,
                        persistence_type="session",
                    ),
                    className="column is-2",
                ),
                html.Div(
                    dcc.Dropdown(
                        id="device-id",
                        placeholder="Device ID",
                        searchable=True,
                        persistence=True,
                        persistence_type="session",
                    ),
                    className="column is-2",
                ),
                html.Div(
                    dcc.Dropdown(
                        id="measurement-type",
                        placeholder="Measurement Type",
                        searchable=True,
                        persistence=True,
                        persistence_type="session",
                    ),
                    className="column is-2",
                ),
                html.Div(
                    dcc.Dropdown(
                        id="measurement",
                        placeholder="Measurement",
                        searchable=True,
                    ),
                    className="column is-2",
                ),
                # Buttons
                html.Div(
                    [
                        html.Button(
                            html.I(className="fa-solid fa-arrow-left"),
                            id="prev-mmnt-button",
                            n_clicks=0,
                            className="button",
                        ),
                        html.Button(
                            html.I(className="fa-solid fa-arrow-right"),
                            id="next-mmnt-button",
                            n_clicks=0,
                            className="button ml-2",
                        ),
                    ],
                    className="column is-1 is-flex is-align-items-center is-justify-content-flex-start",
                ),
            ],
            className="columns is-multiline is-vcentered",
        )


class XarrayData(Data):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @staticmethod
    @callback(
        Output(
            "load-signal",
            "data",
            allow_duplicate=True,
        ),
        State("session-id", "data"),
        State("load-signal", "data"),
        Input("dataset-path", "value"),
        Input("upload-ticker", "n_intervals"),
        prevent_initial_call=True,
    )
    def update_data(sess_id: str, sig_curr: int, path: str, *_):
        """
        Update the data store with the Zarr file content.

        Args:
            sess_id (str): Session ID to load the state for
            sig_curr (int): The current signal value
            path (str): Path to the Zarr file

        Returns:
            dict: The updated data store

        Raises:
            PreventUpdate: If the path is not a file or not a Zarr file

        """
        if path in ["", None]:
            raise PreventUpdate

        path = Path(path)
        if _should_update_data(sess_id, path, sig_curr, "XarrayData"):
            return sig_curr + 1 if sig_curr is not None else 1
        else:
            raise PreventUpdate



import xarray as xr
from pathlib import Path
from dash.exceptions import PreventUpdate

# Local imports
from qimchi.state import load_state_from_disk
from qimchi.logger import logger


def _retry_reader(path, retries=3):
    """
    Utility function to retry reading a dataset

    Args:
        path (Path): Path to the dataset
        retries (int, optional): Number of retries. Defaults to 3.

    Returns:
        xr.Dataset: Loaded xarray dataset

    """
    for i in range(retries):
        try:
            ds = xr.load_dataset(path, engine="zarr")
            return ds
        except Exception as err:
            logger.warning(f"_retry_reader | Error reading: {path} | Err: {err}")
            logger.warning(f"_retry_reader | Retrying... {i + 1}/{retries}")

    logger.error(f"_retry_reader | Failed to read: {path} after {retries} retries")
    raise PreventUpdate


def read_data(sess_id: str, src: str) -> xr.Dataset:
    """
    Utility function to read data into a xarray.Dataset

    Args:
        sess_id (str): Session ID to load the state for
        src (str): Source of the read_data call. For debugging purposes.

    Returns:
        xr.Dataset: Loaded xarray dataset

    """
    _state = load_state_from_disk(sess_id)

    path = Path(_state.measurement_path)
    logger.debug(f"{src} -> read_data | Reading: {path}")

    if not path.is_dir() or not path.suffix == ".zarr":
        logger.warning(
            f"{src} -> read_data | Path is not a directory or not a zarr file: {path}"
        )
        raise PreventUpdate

    try:
        ds = xr.load_dataset(path, engine="zarr")
        logger.debug(f"{src} -> read_data | Reading done: {path}")
        return ds
    # Catch "FileNotFoundError: Unable to find group" and keep retrying
    except FileNotFoundError as err:
        logger.warning(
            f"{src} -> read_data | FileNotFoundError reading: {path} | err: {err}"
        )
        logger.warning(f"{src} -> read_data | Retrying...")
        ds = _retry_reader(path)
        return ds
    except KeyError as err:
        logger.warning(f"{src} -> read_data | KeyError reading: {path} | err: {err}")
        logger.warning(f"{src} -> read_data | Retrying...")
        ds = _retry_reader(path)
        return ds
    except Exception as err:
        logger.error(
            f"{src} -> read_data | Error reading: {path} | err: {err}", exc_info=True
        )
        raise PreventUpdate


def format_number(value: float, default_precision: float = 3) -> str:
    """
    Format a number with scientific notation if it is too small or too large.
    Otherwise, format it with a default precision.

    Args:
        value (float): The number to format.
        default_precision (int): The default number of decimal places to use.

    Returns:
        str: The formatted number as a string.

    """
    if abs(value) < 0.001 or abs(value) >= 10000:
        return f"{value:.3e}"
    else:
        precision = default_precision
        return f"{value:.{precision}f}"



import json
import numpy as np
from uuid import uuid4
from datetime import datetime

import dash_daq as daq
import dash_bootstrap_components as dbc
from dash import dcc, html
from xarray import Dataset
from plotly import graph_objects as go

# Local imports
from qimchi.components.figures import QimchiFigure, HeatMap, Line
from qimchi.components.filters import apply_filters
from qimchi.components.plot_elements import (
    appearance_plot_width_dropdown,
    # appearance_axes_ranges,  # TODOLATER:
    appearance_line_dropdown,
    appearance_line_options,
    appearance_marker_options,
)
from qimchi.components.plot_filters import content_filters

from qimchi.logger import logger
from qimchi.state import (
    load_state_from_disk,
    # Plot size options
    DEFAULT_PLOT_WIDTH,
    SQUARIFY_SIZE,
    # Hmap options
    DEFAULT_HMAP_COLSCALE,
    # Line & marker options
    DEFAULT_LINE_MODE,
    DEFAULT_LINE_COLOR,
    DEFAULT_LINE_WIDTH,
    DEFAULT_LINE_OPACITY,
    DEFAULT_LINE_DASH,
    DEFAULT_LINE_SHAPE,
    DEFAULT_SPLINE_SMOOTHING,
    DEFAULT_MARKER_COLOR,
    DEFAULT_MARKER_SIZE,
    DEFAULT_MARKER_SYMBOL,
    DEFAULT_MARKER_OPACITY,
    # Grid & tick options
    DEFAULT_AXIS_TYPE,
    DEFAULT_GRID_COLOR,
    DEFAULT_TICK_COLOR,
    DEFAULT_SHOWGRID,
    DEFAULT_NTICKS,
    DEFAULT_GRID_WIDTH,
    DEFAULT_GRID_DASH,
    DEFAULT_TICK_LENGTH,
    DEFAULT_TICK_ANGLE,
    DEFAULT_TICK_WIDTH,
    AXIS_TYPE_OPTS,
    GRID_COLOR_OPTS,
    GRID_DASH_OPTS,
    TICK_COLOR_OPTS,
)


DEFAULT_THEME = {
    # NOTE: Nesting structure is NOT indicative of the actual data structure in Plotly figures
    "hmap": {
        "colorscale": DEFAULT_HMAP_COLSCALE,
        "rangecolor": None,  # NOTE: Dynamically set via callback
    },
    "line": {
        "mode": DEFAULT_LINE_MODE,
        "color": DEFAULT_LINE_COLOR,
        "width": DEFAULT_LINE_WIDTH,
        "opacity": DEFAULT_LINE_OPACITY,
        "dash": DEFAULT_LINE_DASH,
        "shape": DEFAULT_LINE_SHAPE,
        "smoothing": DEFAULT_SPLINE_SMOOTHING,
    },
    "marker": {
        "color": DEFAULT_MARKER_COLOR,
        "size": DEFAULT_MARKER_SIZE,
        "symbol": DEFAULT_MARKER_SYMBOL,
        "opacity": DEFAULT_MARKER_OPACITY,
    },
    "x": {
        "maj": {
            "showgrid": DEFAULT_SHOWGRID,
            "type": DEFAULT_AXIS_TYPE,
            "nticks": DEFAULT_NTICKS,
            "gridcolor": DEFAULT_GRID_COLOR,
            "griddash": DEFAULT_GRID_DASH,
            "gridwidth": DEFAULT_GRID_WIDTH,
            "tickcolor": DEFAULT_TICK_COLOR,
            "tickwidth": DEFAULT_TICK_WIDTH,
            "ticklen": DEFAULT_TICK_LENGTH,
            "tickangle": DEFAULT_TICK_ANGLE,
        },
        "min": {
            "showgrid": DEFAULT_SHOWGRID,
            "nticks": DEFAULT_NTICKS,
            "gridcolor": DEFAULT_GRID_COLOR,
            "griddash": DEFAULT_GRID_DASH,
            "gridwidth": DEFAULT_GRID_WIDTH,
            "tickcolor": DEFAULT_TICK_COLOR,
            "tickwidth": DEFAULT_TICK_WIDTH,
            "ticklen": DEFAULT_TICK_LENGTH,
        },
    },
    "y": {
        "maj": {
            "showgrid": DEFAULT_SHOWGRID,
            "type": DEFAULT_AXIS_TYPE,
            "nticks": DEFAULT_NTICKS,
            "gridcolor": DEFAULT_GRID_COLOR,
            "griddash": DEFAULT_GRID_DASH,
            "gridwidth": DEFAULT_GRID_WIDTH,
            "tickcolor": DEFAULT_TICK_COLOR,
            "tickwidth": DEFAULT_TICK_WIDTH,
            "ticklen": DEFAULT_TICK_LENGTH,
            "tickangle": DEFAULT_TICK_ANGLE,
        },
        "min": {
            "showgrid": DEFAULT_SHOWGRID,
            "nticks": DEFAULT_NTICKS,
            "gridcolor": DEFAULT_GRID_COLOR,
            "griddash": DEFAULT_GRID_DASH,
            "gridwidth": DEFAULT_GRID_WIDTH,
            "tickcolor": DEFAULT_TICK_COLOR,
            "tickwidth": DEFAULT_TICK_WIDTH,
            "ticklen": DEFAULT_TICK_LENGTH,
        },
    },
}


DEFAULT_PLOTS_MENU = {
    "ranges_n_clicks": 0,
    "ranges_cname": "button mb-2",
    "appearance_n_clicks": 0,
    "appearance_cname": "button mb-2",
    "filters_n_clicks": 0,
    "filters_cname": "button mb-2",
    "plots_menu_dropdown_disp": "none",
    "content_ranges_disp": "none",
    "content_appearance_disp": "none",
    "content_filters_disp": "none",
    "appearance_line_options_disp": True,
    "appearance_marker_options_disp": True,
}


def content_ranges(
    sess_id: str, slider: dict, fig: QimchiFigure, fig_id: str
) -> html.Div:
    """
    Generates the range sliders for the plot.

    Args:
        sess_id (str): Session ID to load the state for
        slider (dict): Slider configuration
        fig (QimchiFigure): QimchiFigure object.
        fig_id (str): Unique identifier for the figure in the plot

    Returns:
        dash.html.Div: Range sliders container.

    """
    _state = load_state_from_disk(sess_id)

    if fig.num_axes == 2:
        rcol_min = float(np.nanmin(fig.data))
        rcol_max = float(np.nanmax(fig.data))
        fig_state = _state.plot_states[fig_id]
        rangecolor = fig_state["theme"]["hmap"]["rangecolor"]
        # logger.debug(f"Initial rangecolor: {rangecolor}")

        if rangecolor is not None:
            rangecolor[0] = max(float(rangecolor[0]), rcol_min)
            rangecolor[1] = min(float(rangecolor[1]), rcol_max)
        else:
            rangecolor = [rcol_min, rcol_max]
        fig_state["theme"]["hmap"]["rangecolor"] = rangecolor
        _state.save_state()
        # logger.debug(f"Updated rangecolor: {rangecolor}")
        rcol_marks = {
            rcol_min: f"{rcol_min:.3f}",
            rcol_max: f"{rcol_max:.3f}",
        }

    fig_plot_menu = _state.plot_states[fig_id]["plots_menu"]
    dim_sliders = []
    for dim in slider:
        dim_sliders.append(
            html.Div(
                [
                    # 0 padding and negative margin to align the label with the slider
                    html.Div(
                        f"{dim} : ",
                        className="column is-1 p-0",
                        style={"fontWeight": "bold", "margin-top": "-0.3225rem"},
                    ),
                    dcc.Slider(
                        min=slider[dim]["min"],
                        max=slider[dim]["max"],
                        value=slider[dim]["value"],
                        marks=None,
                        updatemode="drag",
                        tooltip={
                            "placement": "bottom",
                            "always_visible": True,
                        },
                        className="column is-11",
                        id={
                            "type": "slider",
                            "index": fig_id,
                            "menu": "ranges",
                            "extra": json.dumps(
                                {
                                    "dim": dim,
                                }
                            ),
                        },
                    ),
                ],
                className="columns is-full",
            )
        )

    dim_sliders.append(
        html.Div(
            [
                dcc.Input(
                    id={
                        "index": fig_id,
                        "key": "hmap",
                        "setting": "textrcolmin",
                    },
                    type="number",
                    placeholder="Min",
                    inputMode="numeric",
                    className="input",
                    style={
                        "width": "45%",
                        "margin": "0.3rem 0.2rem 0 0",
                    },
                )
                if fig.num_axes == 2
                else None,
                dcc.Input(
                    id={
                        "index": fig_id,
                        "key": "hmap",
                        "setting": "textrcolmax",
                    },
                    type="number",
                    placeholder="Max",
                    inputMode="numeric",
                    className="input",
                    style={
                        "width": "45%",
                        "margin": "0.3rem 0 0 0.2rem",
                    },
                )
                if fig.num_axes == 2
                else None,
                html.Button(
                    html.I(className="fa-solid fa-paper-plane"),
                    id={
                        "index": fig_id,
                        "key": "hmap",
                        "setting": "textrcolapply",
                    },
                    className="button is-pulled-right",
                    style={
                        "width": "7.5%",
                        "padding": "0.65rem 1rem",
                    },
                )
                if fig.num_axes == 2
                else None,
                dcc.RangeSlider(
                    min=rcol_min,
                    max=rcol_max,
                    marks=rcol_marks,
                    value=rangecolor,
                    updatemode="drag",
                    id={
                        "index": fig_id,
                        "key": "hmap",
                        "setting": "rangecolor",
                    },
                    tooltip={"placement": "bottom"},
                    persistence=True,
                    persistence_type="local",
                    className="mx-0 my-3 px-0",
                )
                if fig.num_axes == 2
                else None,
            ],
            className="p-1",
        )
    )

    return html.Div(
        dim_sliders,
        className="column is-full",
        style={"display": fig_plot_menu["content_ranges_disp"]},
        id={"type": "content-ranges", "index": fig_id},
    )


def _grid_options(fig: QimchiFigure, fig_id: str, axis: str, ticks: str) -> html.Div:
    """
    Generates the grid options for the plot.

    Args:
        fig (QimchiFigure): QimchiFigure object.
        fig_id (str): Unique identifier for the figure in the plot.
        axis (str): Axis to apply the grid options to. Either "x" or "y".
        ticks (str): Major or minor ticks. Either "maj" or "min".

    Returns:
        dash.html.Div: Grid options container.

    """
    # Flag to disable certain options for HeatMap
    is_hmap = fig.__class__.__name__ == "HeatMap"
    # Flag to disable certain options for minor ticks
    is_min = ticks == "min"

    return html.Div(
        [
            # Tick Options
            html.Div(
                [
                    html.Label(
                        "Tick Options",
                        className="column is-full",
                        style={"fontWeight": "bold"},
                    ),
                    dcc.Dropdown(
                        id={
                            "index": fig_id,
                            "axis": axis,
                            "ticks": ticks,
                            "setting": "tickcolor",
                        },
                        options=TICK_COLOR_OPTS,
                        value=DEFAULT_TICK_COLOR,
                        placeholder="Tick Color",
                        searchable=True,
                        persistence=True,
                        persistence_type="local",
                        className="column is-full",
                    ),
                    html.Div(
                        [
                            html.Label(
                                "Tick Length: ",
                                className="column is-one-fifth",
                                style={"fontWeight": "bold"},
                            ),
                            dcc.Slider(
                                min=0.0,
                                max=50.0,
                                step=0.5,
                                marks={0: "0", 50: "50"},
                                updatemode="drag",
                                tooltip={
                                    "placement": "bottom",
                                    "always_visible": True,
                                },
                                value=DEFAULT_TICK_LENGTH,
                                id={
                                    "index": fig_id,
                                    "axis": axis,
                                    "ticks": ticks,
                                    "setting": "ticklen",
                                },
                                persistence=True,
                                persistence_type="local",
                                className="column is-four-fifths mx-0 my-2",
                            ),
                        ],
                        className="column is-flex is-one-half",
                    ),
                    html.Div(
                        [
                            html.Label(
                                "Tick Width: ",
                                className="column is-one-fifth",
                                style={"fontWeight": "bold"},
                            ),
                            dcc.Slider(
                                min=0.0,
                                max=15.0,
                                step=0.5,
                                marks={0: "0", 15: "15"},
                                value=DEFAULT_TICK_WIDTH,
                                updatemode="drag",
                                tooltip={
                                    "placement": "bottom",
                                    "always_visible": True,
                                },
                                id={
                                    "index": fig_id,
                                    "axis": axis,
                                    "ticks": ticks,
                                    "setting": "tickwidth",
                                },
                                persistence=True,
                                persistence_type="local",
                                className="column is-four-fifths mx-0 my-2",
                            ),
                        ],
                        className="column is-flex is-one-half",
                    ),
                    (
                        html.Div(
                            [
                                html.Label(
                                    "Tick Angle: ",
                                    className="column is-one-fifth",
                                    style={"fontWeight": "bold"},
                                ),
                                dcc.Slider(
                                    min=-90.0,
                                    max=90.0,
                                    step=5.0,
                                    marks={
                                        -90: "-90",
                                        -45: "-45",
                                        0: "0",
                                        45: "45",
                                        90: "90",
                                    },
                                    value=DEFAULT_TICK_ANGLE,
                                    id={
                                        "index": fig_id,
                                        "axis": axis,
                                        "ticks": ticks,
                                        "setting": "tickangle",
                                    },
                                    updatemode="drag",
                                    tooltip={
                                        "placement": "bottom",
                                        "always_visible": True,
                                    },
                                    persistence=True,
                                    persistence_type="local",
                                    className="column is-four-fifths mx-0 my-2",
                                ),
                            ],
                            className="column is-flex is-one-half",
                        )
                        if not is_min
                        else None
                    ),
                ],
                className="column is-half px-0",
            ),
            # Grid Options
            (
                html.Div(
                    [
                        html.Div(
                            [
                                html.Label(
                                    "Grid Options",
                                    className="column is-four-fifths",
                                    style={"fontWeight": "bold"},
                                ),
                                daq.ToggleSwitch(
                                    id={
                                        "index": fig_id,
                                        "axis": axis,
                                        "ticks": ticks,
                                        "setting": "showgrid",
                                    },
                                    value=DEFAULT_SHOWGRID,
                                    labelPosition="top",
                                    persistence=True,
                                    persistence_type="local",
                                    className="column is-one-fifth",
                                ),
                            ],
                            className="column is-full is-flex p-0",
                        ),
                        (
                            dcc.Dropdown(
                                id={
                                    "index": fig_id,
                                    "axis": axis,
                                    "ticks": ticks,
                                    "setting": "type",
                                },
                                options=AXIS_TYPE_OPTS,
                                value=DEFAULT_AXIS_TYPE,
                                placeholder="Axis Type",
                                searchable=True,
                                persistence=True,
                                persistence_type="local",
                                className="column is-full",
                            )
                            if not is_min
                            else None
                        ),
                        dcc.Dropdown(
                            id={
                                "index": fig_id,
                                "axis": axis,
                                "ticks": ticks,
                                "setting": "gridcolor",
                            },
                            options=GRID_COLOR_OPTS,
                            value=DEFAULT_GRID_COLOR,
                            placeholder="Grid Color",
                            searchable=True,
                            persistence=True,
                            persistence_type="local",
                            className="column is-full",
                        ),
                        dcc.Dropdown(
                            id={
                                "index": fig_id,
                                "axis": axis,
                                "ticks": ticks,
                                "setting": "griddash",
                            },
                            options=GRID_DASH_OPTS,
                            value=DEFAULT_GRID_DASH,
                            placeholder="Grid Dash",
                            clearable=False,
                            persistence=True,
                            persistence_type="local",
                            className="column is-full",
                        ),
                        html.Div(
                            [
                                html.Label(
                                    "Num Ticks: ",
                                    className="column is-one-fifth",
                                    style={"fontWeight": "bold"},
                                ),
                                dcc.Slider(
                                    min=0,
                                    max=100,
                                    step=1,
                                    marks={0: "0", 100: "100"},
                                    value=DEFAULT_NTICKS,
                                    id={
                                        "index": fig_id,
                                        "axis": axis,
                                        "ticks": ticks,
                                        "setting": "nticks",
                                    },
                                    updatemode="drag",
                                    tooltip={
                                        "placement": "bottom",
                                        "always_visible": True,
                                    },
                                    persistence=True,
                                    persistence_type="local",
                                    className="column is-four-fifths mx-0 my-2",
                                ),
                            ],
                            className="column is-flex is-one-half",
                        ),
                        html.Div(
                            [
                                html.Label(
                                    "Grid Width: ",
                                    className="column is-one-fifth",
                                    style={"fontWeight": "bold"},
                                ),
                                dcc.Slider(
                                    min=0.0,
                                    max=10.0,
                                    step=0.5,
                                    marks={0: "0", 10: "10"},
                                    value=DEFAULT_GRID_WIDTH,
                                    id={
                                        "index": fig_id,
                                        "axis": axis,
                                        "ticks": ticks,
                                        "setting": "gridwidth",
                                    },
                                    updatemode="drag",
                                    tooltip={
                                        "placement": "bottom",
                                        "always_visible": True,
                                    },
                                    persistence=True,
                                    persistence_type="local",
                                    className="column is-four-fifths mx-0 my-2",
                                ),
                            ],
                            className="column is-flex is-one-half",
                        ),
                    ],
                    className="column is-half px-0",
                )
                if not is_hmap
                else None
            ),
        ],
        className="column is-full is-flex",
    )


def content_appearance(sess_id: str, fig: QimchiFigure, fig_id: str) -> html.Div:
    """
    Generates the appearance menu for the plot.

    Args:
        sess_id (str): Session ID to load the state for
        fig (QimchiFigure): QimchiFigure object
        fig_id (str): Unique identifier for the figure in the plot

    Returns:
        dash.html.Div: Appearance menu container.

    """
    _state = load_state_from_disk(sess_id)

    logger.debug(
        f"content_appearance() | fig_id: {fig_id} | fig.num_axes: {fig.num_axes}"
    )

    fig_plot_menu = _state.plot_states[fig_id]["plots_menu"]
    general_options = dcc.Tabs(
        [
            dcc.Tab(
                label="X-Axis Major Ticks & Grids",
                className="p-1",
                children=[_grid_options(fig, fig_id, "x", "maj")],
            ),
            dcc.Tab(
                label="X-Axis Minor Ticks & Grids",
                className="p-1",
                children=[_grid_options(fig, fig_id, "x", "min")],
            ),
            dcc.Tab(
                label="Y-Axis Major Ticks & Grids",
                className="p-1",
                children=[_grid_options(fig, fig_id, "y", "maj")],
            ),
            dcc.Tab(
                label="Y-Axis Minor Ticks & Grids",
                className="p-1",
                children=[_grid_options(fig, fig_id, "y", "min")],
            ),
        ]
    )

    # For Heat Maps
    hmap_options = None
    if fig.num_axes == 2:
        colors = fig.colors

        hmap_options = html.Div(
            [
                appearance_plot_width_dropdown(fig_id),
                dcc.Dropdown(
                    options=colors,
                    id={
                        "index": fig_id,
                        "key": "hmap",
                        "setting": "colorscale",
                    },
                    value=DEFAULT_HMAP_COLSCALE,
                    placeholder="Color Scale",
                    searchable=True,
                    persistence=True,
                    persistence_type="local",
                ),
            ]
        )

    # For Line/Scatter plots
    line_options = None
    if fig.num_axes == 1:
        line_options = html.Div(
            children=[
                appearance_plot_width_dropdown(fig_id),
                # appearance_axes_ranges(fig_id), # TODOLATER:
                appearance_line_dropdown(fig_id),
                html.Div(
                    id={
                        "index": fig_id,
                        "type": "line-options",
                    },
                    children=[
                        (
                            appearance_line_options(fig_id)
                            if fig_plot_menu["appearance_line_options_disp"]
                            else None
                        ),
                        (
                            appearance_marker_options(fig_id)
                            if fig_plot_menu["appearance_marker_options_disp"]
                            else None
                        ),
                    ],
                ),
            ],
        )

    fig_cname = fig.__class__.__name__
    plot_options = dcc.Tab(
        label=f"{fig_cname} Settings",
        className="p-1",
        children=[hmap_options if fig_cname == "HeatMap" else line_options],
    )

    return html.Div(
        [
            dcc.Tabs(
                [
                    plot_options,
                    dcc.Tab(
                        label="General Settings",
                        className="p-1",
                        children=[general_options],
                    ),
                ],
                className="column is-full px-0 py-2",
            )
        ],
        style={"display": fig_plot_menu["content_appearance_disp"]},
        id={"type": "content-appearance", "index": fig_id},
    )


def plots_menu(sess_id: str, fig_id: str) -> list[html.Button]:
    """
    Generator for the plots menu, defining the buttons for the plot options.

    Args:
        sess_id (str): Session ID to load the state for
        fig_id (str): Unique value for associating buttons to their parent graphs

    Returns:
        list: List of buttons for the plot menu

    """
    _state = load_state_from_disk(sess_id)

    fig_plot_menu = _state.plot_states[fig_id]["plots_menu"]

    send_to_md = html.Button(
        html.I(className="fa-solid fa-clipboard-list"),
        id={"index": fig_id, "type": "button-plot-send-to-md"},
        className="button",
        disabled=True,  # TODOLATER: Enable when bug is fixed. See notes.py::update_notes()
        title="Send to Markdown",
    )

    export = html.Button(
        html.I(className="fa-solid fa-save"),
        id={"index": fig_id, "type": "button-plot-export"},
        className="button",
        title="Export Plot",
    )

    ranges = html.Button(
        html.I(className="fa-solid fa-arrow-right-arrow-left"),
        id={"index": fig_id, "type": "button-plot-ranges"},
        n_clicks=fig_plot_menu["ranges_n_clicks"],
        className=fig_plot_menu["ranges_cname"],
        title="Edit Plot Range",
    )

    appearance = html.Button(
        html.I(className="fa-solid fa-palette"),
        id={"index": fig_id, "type": "button-plot-appearance"},
        n_clicks=fig_plot_menu["appearance_n_clicks"],
        className=fig_plot_menu["appearance_cname"],
        title="Edit Plot Appearance",
    )

    filters = html.Button(
        html.I(className="fa-solid fa-meteor"),
        id={"index": fig_id, "type": "button-plot-filters"},
        n_clicks=fig_plot_menu["filters_n_clicks"],
        className=fig_plot_menu["filters_cname"],
        title="Apply Filters",
    )

    delete = html.Button(
        html.I(className="fa-solid fa-xmark"),
        id={"index": fig_id, "type": "button-plot-delete"},
        n_clicks=0,
        className="button is-danger mb-2",
        title="Delete Plot",
    )

    return [
        send_to_md,
        export,
        ranges,
        appearance,
        filters,
        delete,
    ]


def plots_menu_dropdown(sess_id: str, dropdown_options: list, fig_id: str) -> html.Div:
    """
    Dropdown container for each of the plot menu options.
    Normally hidden until the corresponding button is clicked.

    Args:
        sess_id (str): Session ID to load the state for
        dropdown_options (list): List of Buttons inside the dropdown
        fig_id (str): Unique value for associating dropdowns to their parent graphs

    Returns:
        dash.html.Div: Dropdown container

    """
    _state = load_state_from_disk(sess_id)

    fig_plot_menu = _state.plot_states[fig_id]["plots_menu"]

    return html.Div(
        html.Div(
            dropdown_options,
            className="column is-10",
            style={"display": fig_plot_menu["plots_menu_dropdown_disp"]},
            id={"type": "plots-menu-dropdown", "index": fig_id},
        ),
        className="is-flex is-justify-content-center is-align-items-center",
        style={"width": "100%"},
    )


class Plot:
    def __init__(
        self,
        sess_id: str,
        data: Dataset,
        deps: list,
        indeps: list,
        fig_id: str = None,
        slider: dict = {},
        theme: dict = DEFAULT_THEME,
        filters_order: list = [],
        filters_opts: dict = {},
        plots_menu: dict = DEFAULT_PLOTS_MENU,
        width_class: str = DEFAULT_PLOT_WIDTH,
    ):
        """
        Main plot class; generates the plot based on the data, dependent and independent variables.

        Args:
            sess_id (str): Session ID to load the state for.
            data (xarray.Dataset): Xarray data loaded from the selected file.
            deps (list): Main dependent variable(s) to plot.
            indeps (list): Independent variables to plot.
            fig_id (str): Unique identifier for the figure in the plot.
            slider (dict, optional): Slider configuration. Defaults to {}.
            theme (dict, optional): Plot theme dictionary. Defaults to `DEFAULT_THEME`.
            filters_order (list, optional): Filters in order. Defaults to [].
            filters_opts (dict, optional): Filters options. Defaults to {}. If a filter has no options, it is {}.
            plots_menu (dict, optional): Plot menu configuration. Defaults to `DEFAULT_PLOTS_MENU`.
            width_class (str, optional): Plot width Bulma className. Defaults to `DEFAULT_PLOT_WIDTH`.

        Raises:
            ValueError: If `indeps` or `deps` are not lists.

        """
        if not isinstance(indeps, list):
            err = "Independents must be a list"
            # logger.error(err)
            raise ValueError(err)
        if not isinstance(deps, list):
            err = "Dependents must be a list"
            # logger.error(err)
            raise ValueError(err)

        # TODOLATER: When deps is a proper list with more than 1 item, remove this hardcoding
        deps = deps[0]
        self.data_array = data[deps]

        # NOTE: These are stored in _state
        self.id = str(uuid4()) if not fig_id else fig_id
        self.indeps = indeps
        self.deps = deps
        self.theme = theme
        self.slider = slider
        self.filters_order = filters_order
        self.filters_opts = filters_opts
        self.plots_menu = plots_menu
        self.width_class = width_class

        # State
        self.sess_id = sess_id
        self._state = load_state_from_disk(sess_id)

        # logger.debug(f"Plotting {self.deps} against {self.indeps} with id {self.id}")

    def _base(self) -> QimchiFigure:
        """
        Base Plotly graph generator

        Returns:
            QimchiFigure: Base Figure object of type Line or HeatMap

        Raises:
            ValueError: If the number of independent variables is greater than the number of axes.

        """

        def _update_slider():
            for dim in self.data_array.dims:
                # logger.debug(f"self.indeps | {self.indeps}")
                if dim not in self.indeps:
                    if self.slider == {} or dim not in self.slider.keys():
                        # logger.debug("Slider is empty")
                        vals = self.data_array.coords[dim].values
                        self.slider[dim] = {
                            "min": min(vals),
                            "max": max(vals),
                            "step": vals[1] - vals[0],
                            "value": min(vals),
                        }

        # Line Plot
        if len(self.indeps) == 1:
            # logger.debug(f"self.data_array.dims | {self.data_array.dims}")
            if len(self.data_array.dims) > 1:
                _update_slider()
            slider_vals = {dim: self.slider[dim]["value"] for dim in self.slider}
            data_array_subset = self.data_array.sel(**slider_vals, method="nearest")
            fig = Line(
                self.sess_id, data_array_subset, self.indeps, self.deps, self.theme
            )

        # Heat Map
        elif len(self.indeps) == 2:
            if len(self.data_array.dims) > 2:
                _update_slider()
            slider_vals = {dim: self.slider[dim]["value"] for dim in self.slider}
            data_array_subset = self.data_array.sel(**slider_vals, method="nearest")
            fig = HeatMap(
                self.sess_id, data_array_subset, self.indeps, self.deps, self.theme
            )

        else:
            err = f"Invalid number of independent variables: {len(self.indeps)}"
            # logger.error(err)
            raise ValueError(err)

        return fig

    def plot(self, return_div: bool = True) -> html.Div | go.Figure:
        """
        Creates the plot container with the plot & dropdowns, if `return_div` is True.
        Only the plot is returned, if `return_div` is False.

        Args:
            return_div (bool, optional): Whether to return the plot container or just the plot (go.Figure). Defaults to True.

        Returns:
            go.Figure | dash.html.Div: Either the plot or the plot container, depending on `return_div`.

        """
        fig: QimchiFigure = self._base()
        fig_plot: go.Figure = fig.plot()
        ID = self.id
        sess_id = self.sess_id
        _state = self._state

        _state.plot_states.update(
            {
                ID: {
                    "plot_type": fig.__class__.__name__,
                    "deps": [self.deps],  # TODOLATER: when deps is a list
                    "indeps": self.indeps,
                    # NOTE: Deep copy of theme
                    "theme": json.loads(json.dumps(self.theme)),
                    "slider": self.slider,
                    "filters_order": self.filters_order,
                    "filters_opts": self.filters_opts,
                    # NOTE: Deep copy of plots_menu
                    "plots_menu": json.loads(json.dumps(self.plots_menu)),
                    "width_class": self.width_class,
                }
            }
        )
        _state.save_state()

        logger.debug(
            f"Plot.plot() | Applying filters with new filter_opts: {self.filters_opts}"
        )
        if self.filters_order:
            fig_plot = apply_filters(
                filters_order=self.filters_order,
                filters_opts=self.filters_opts,
                fig=fig_plot,
                fig_num_axes=fig.num_axes,
            )

        if fig.__class__.__name__ == "HeatMap":
            fig.data = fig_plot.data[0]["z"]
            logger.debug(
                "Plot.plot() | Updated HeatMap data for rangecolor RangeSlider."
            )

        if not return_div:
            return fig_plot

        fig_cls = fig.__class__.__name__
        # is_line = fig_cls == "Line"
        style = {
            "width": SQUARIFY_SIZE if _state.squarify_plots else "auto",
            "height": SQUARIFY_SIZE if _state.squarify_plots else "auto",
        }
        fig_graph = dcc.Graph(
            figure=fig_plot,
            id={"index": ID, "type": "figure"},
            style=style,
            mathjax=True,
            # To unset the last hover label to hide the custom-hover Div when exiting hover
            clear_on_unhover=True,
            # NOTE: Animation is only for Line plots
            # BUG: Does not animate axis ranges, meaning charts can drop off the plot
            # BUG: See: https://github.com/plotly/plotly.js/issues/1687 (Probably #wontfix)
            # animate=is_line,
            # animation_options={
            #     "frame": {"redraw": False},
            #     "transition": {"duration": 500, "easing": "cubic-in-out"},
            # }
            # if is_line
            # else {},
            config={
                "editable": True,
                "edits": {
                    # TODOLATER: if needed
                    # "legendPosition": True,
                    # "legendText": True,
                    # "shapePosition": True,
                    "annotationTail": True,
                    "annotationText": False,
                    "annotationPosition": True,
                    "axisTitleText": True,
                    "colorbarTitleText": True,
                    "colorbarPosition": True,
                    "titleText": False,
                },
                "displaylogo": False,  # Hides the Plotly logo
                "toImageButtonOptions": {
                    "format": "svg",  # One of png, svg, jpeg, webp
                    "filename": f"{self.id[:8]}__{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}__{fig_cls}",
                    # TODOLATER: if needed
                    # "width": None,
                    # "height": None,
                    # "scale": 1,
                },
                # TODOLATER: if needed
                # "modeBarButtonsToAdd": [],
                # "modeBarButtonsToRemove": [],
            },
        )
        return html.Div(
            [
                html.Div(
                    [
                        # Custom hover label
                        html.Div(
                            id={"index": ID, "type": "custom-hover"},
                            style={
                                "fontFamily": "monospace",
                                "textAlign": "center",
                                "marginBottom": "4px",
                                "fontWeight": "bold",
                                "fontSize": "10px",
                                "height": "12px",
                            },
                        ),
                        fig_graph,
                        # Export download
                        dcc.Download(id={"index": ID, "type": "export-download"}),
                        # Store to save relayout information
                        dcc.Store(
                            id={
                                "index": ID,
                                "type": "relayout-data",
                            },
                            storage_type="local",
                        ),
                        # Store to save notes
                        dcc.Store(
                            id={
                                "index": ID,
                                "type": "notes-store",
                            },
                            storage_type="local",
                        ),
                        # Toast to show export status
                        dbc.Toast(
                            "",
                            id={"index": ID, "type": "export-toast"},
                            header=None,
                            header_style={"display": "none"},
                            is_open=False,
                            dismissable=True,
                            duration=3000,
                            # NOTE: Styled using Bulma, not Bootstrap
                            style={
                                "position": "fixed",
                                "top": 50,
                                "right": 10,
                                "width": 350,
                            },
                        ),
                    ],
                    className="column is-11",
                ),
                html.Div(
                    plots_menu(sess_id, fig_id=ID),
                    className="column is-1 mt-auto",
                ),
                plots_menu_dropdown(
                    sess_id=sess_id,
                    dropdown_options=[
                        content_ranges(sess_id, self.slider, fig, fig_id=ID),
                        content_appearance(sess_id, fig, fig_id=ID),
                        content_filters(sess_id, fig, fig_id=ID),
                    ],
                    fig_id=ID,
                ),
            ],
            className="columns is-multiline is-full",
        )



import os
import logging
from rich.logging import RichHandler


# Get debug level from environment variable
debug_level = int(os.environ.get("QIMCHI_DEBUG_LEVEL", logging.INFO))

logging.basicConfig(
    level=debug_level,
    format="%(message)s",
    handlers=[RichHandler()],
)

# Qimchi Logger
logger = logging.getLogger(__name__)
